# -*- coding: utf-8 -*-
from __future__ import annotations

import ctypes
import logging
import platform
from typing import cast

from packaging.version import Version
from rubicon.objc import ObjCClass
from rubicon.objc.runtime import load_library

logger = logging.getLogger(__name__)
macos_version = Version(platform.mac_ver()[0])


__all__ = [
    "is_bundle",
    "is_signed_bundle",
    "macos_version",
]


sec = load_library("Security")
foundation = load_library("Foundation")

NSBundle = ObjCClass("NSBundle")

# Security/SecRequirement.h

kSecCSDefaultFlags = 0
kSecCSCheckAllArchitectures = 1 << 0
kSecCSDoNotValidateExecutable = 1 << 1
kSecCSDoNotValidateResources = 1 << 2
kSecCSCheckNestedCode = 1 << 3
kSecCSStrictValidate = 1 << 4


def is_bundle() -> bool:
    """
    Detect if we are in an app bundle

    :returns: Whether we are inside an app bundle.
    """
    return NSBundle.mainBundle.bundleIdentifier is not None


def is_signed_bundle() -> bool:
    """
    Detect if we are in a signed app bundle

    :returns: Whether we are inside a signed app bundle.
    """
    if not is_bundle():
        return False

    # Check for valid code signature on bundle.
    static_code = ctypes.c_void_p(0)
    err = sec.SecStaticCodeCreateWithPath(
        NSBundle.mainBundle.bundleURL, kSecCSDefaultFlags, ctypes.byref(static_code)
    )

    if err != 0:
        _codesigning_warning("SecStaticCodeCreateWithPath", err)
        return False

    signed_status = sec.SecStaticCodeCheckValidity(
        static_code,
        kSecCSCheckAllArchitectures | kSecCSCheckNestedCode | kSecCSStrictValidate,
        None,
    )

    signed_status = cast(int, signed_status)

    if signed_status == 0:
        return True
    else:
        _codesigning_warning("SecStaticCodeCheckValidity", signed_status)
        return False


def _codesigning_warning(call: str, os_status: int) -> None:
    """Log a warning about a failed code signing check."""
    logger.warning(
        "Cannot verify signature of bundle %s. %s call failed with OSStatus: %s",
        NSBundle.mainBundle.bundleIdentifier,
        call,
        os_status,
    )



# -*- coding: utf-8 -*-
"""
Dummy backend for unsupported platforms
"""
from __future__ import annotations

from ..common import Capability, Notification
from .base import DesktopNotifierBackend


class DummyNotificationCenter(DesktopNotifierBackend):
    """A dummy backend for unsupported platforms"""

    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send notifications.

        :returns: Whether authorisation has been granted.
        """
        return True

    async def has_authorisation(self) -> bool:
        """
        Whether we have authorisation to send notifications.
        """
        return True

    async def _send(self, notification: Notification) -> None:
        pass

    async def _clear(self, identifier: str) -> None:
        pass

    async def _clear_all(self) -> None:
        pass

    async def _get_capabilities(self) -> frozenset[Capability]:
        return frozenset()



import asyncio
import time
from pathlib import Path

import pytest

from desktop_notifier import (
    DEFAULT_SOUND,
    Attachment,
    Button,
    DesktopNotifier,
    Icon,
    ReplyField,
    Sound,
    Urgency,
)


async def wait_for_notifications(
    notifier: DesktopNotifier, notification_count: int = 1, timeout_sec: float = 2.0
) -> None:
    t0 = time.monotonic()

    while time.monotonic() - t0 < timeout_sec:
        if len(await notifier.get_current_notifications()) == notification_count:
            return
        await asyncio.sleep(0.2)

    raise TimeoutError("Timed out while waiting for notifications")


@pytest.mark.asyncio
async def test_request_authorisation(notifier: DesktopNotifier) -> None:
    """
    Authorization should already be grated when setting up the test environment.
    Check that calling request_authorisation returns true.
    """
    if not await notifier.has_authorisation():
        pytest.skip("Not authorised and we cannot request authorisation in test")

    has_authorisation = await notifier.request_authorisation()
    assert has_authorisation


@pytest.mark.asyncio
async def test_send(notifier: DesktopNotifier) -> None:
    notification = await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
        urgency=Urgency.Critical,
        buttons=[
            Button(title="Mark as read", on_pressed=lambda: print("Marked as read"))
        ],
        reply_field=ReplyField(
            title="Reply",
            button_title="Send",
            on_replied=lambda text: print("Brutus replied:", text),
        ),
        on_dispatched=lambda: print("Notification showing"),
        on_clicked=lambda: print("Notification clicked"),
        on_dismissed=lambda: print("Notification dismissed"),
        sound=DEFAULT_SOUND,
        thread="test_notifications",
        timeout=5,
    )
    await wait_for_notifications(notifier)
    assert notification in await notifier.get_current_notifications()


@pytest.mark.asyncio
async def test_icon_name(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar", message="Et tu, Brute?", icon=Icon(name="call-start")
    )


@pytest.mark.asyncio
async def test_icon_path(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar", message="Et tu, Brute?", icon=Icon(path=Path("/blue"))
    )


@pytest.mark.asyncio
async def test_icon_uri(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar", message="Et tu, Brute?", icon=Icon(uri="file:///blue")
    )


@pytest.mark.asyncio
async def test_sound_name(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar", message="Et tu, Brute?", sound=Sound(name="Tink")
    )


@pytest.mark.asyncio
async def test_sound_path(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar", message="Et tu, Brute?", sound=Sound(path=Path("/blue"))
    )


@pytest.mark.asyncio
async def test_sound_uri(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar", message="Et tu, Brute?", sound=Sound(uri="file:///blue")
    )


@pytest.mark.asyncio
async def test_attachment_path(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
        attachment=Attachment(path=Path("/blue")),
    )


@pytest.mark.asyncio
async def test_attachment_uri(notifier: DesktopNotifier) -> None:
    await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
        attachment=Attachment(uri="file:///blue"),
    )


@pytest.mark.asyncio
async def test_clear(notifier: DesktopNotifier) -> None:
    n0 = await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    n1 = await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    await wait_for_notifications(notifier, 2)
    current_notifications = await notifier.get_current_notifications()

    assert n0 in current_notifications
    assert n1 in current_notifications

    await notifier.clear(n0)
    assert n0 not in await notifier.get_current_notifications()


@pytest.mark.asyncio
async def test_clear_all(notifier: DesktopNotifier) -> None:
    n0 = await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    n1 = await notifier.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    await wait_for_notifications(notifier, 2)
    current_notifications = await notifier.get_current_notifications()

    assert n0 in current_notifications
    assert n1 in current_notifications

    await notifier.clear_all()
    assert len(await notifier.get_current_notifications()) == 0



# -*- coding: utf-8 -*-
"""
UNUserNotificationCenter backend for macOS

* Introduced in macOS 10.14.
* Cross-platform with iOS and iPadOS.
* Only available from signed app bundles if called from the main executable or from a
  signed Python framework (for example from python.org).
* Requires a running CFRunLoop to invoke callbacks.
"""
from __future__ import annotations

import asyncio
import enum
import logging
import shutil
import tempfile
from concurrent.futures import Future
from pathlib import Path

from packaging.version import Version
from rubicon.objc import NSObject, ObjCClass, objc_method, py_from_ns
from rubicon.objc.runtime import load_library, objc_block, objc_id

from ..common import DEFAULT_SOUND, Capability, Icon, Notification, Urgency
from .base import DesktopNotifierBackend
from .macos_support import macos_version

__all__ = ["CocoaNotificationCenter"]

logger = logging.getLogger(__name__)

foundation = load_library("Foundation")
uns = load_library("UserNotifications")

UNNotification = ObjCClass("UNNotification")
UNUserNotificationCenter = ObjCClass("UNUserNotificationCenter")
UNMutableNotificationContent = ObjCClass("UNMutableNotificationContent")
UNNotificationRequest = ObjCClass("UNNotificationRequest")
UNNotificationAction = ObjCClass("UNNotificationAction")
UNTextInputNotificationAction = ObjCClass("UNTextInputNotificationAction")
UNNotificationCategory = ObjCClass("UNNotificationCategory")
UNNotificationSound = ObjCClass("UNNotificationSound")
UNNotificationAttachment = ObjCClass("UNNotificationAttachment")
UNNotificationSettings = ObjCClass("UNNotificationSettings")

NSURL = ObjCClass("NSURL")
NSSet = ObjCClass("NSSet")
NSError = ObjCClass("NSError")

# UserNotifications.h

UNNotificationDefaultActionIdentifier = (
    "com.apple.UNNotificationDefaultActionIdentifier"
)
UNNotificationDismissActionIdentifier = (
    "com.apple.UNNotificationDismissActionIdentifier"
)

UNAuthorizationOptionBadge = 1 << 0
UNAuthorizationOptionSound = 1 << 1
UNAuthorizationOptionAlert = 1 << 2

UNNotificationActionOptionAuthenticationRequired = 1 << 0
UNNotificationActionOptionDestructive = 1 << 1
UNNotificationActionOptionForeground = 1 << 2
UNNotificationActionOptionNone = 0

UNNotificationCategoryOptionNone = 0
UNNotificationCategoryOptionCustomDismissAction = 1

UNAuthorizationStatusAuthorized = 2
UNAuthorizationStatusProvisional = 3
UNAuthorizationStatusEphemeral = 4


class UNNotificationInterruptionLevel(enum.Enum):
    Passive = 0
    Active = 1
    TimeSensitive = 2
    Critical = 3


ReplyActionIdentifier = "com.desktop-notifier.ReplyActionIdentifier"


class NotificationCenterDelegate(NSObject):  # type:ignore[misc]
    """Delegate to handle user interactions with notifications"""

    implementation: CocoaNotificationCenter

    @objc_method  # type:ignore[misc]
    def userNotificationCenter_didReceiveNotificationResponse_withCompletionHandler_(
        self, center: objc_id, response: objc_id, completion_handler: objc_block
    ) -> None:
        identifier = py_from_ns(response.notification.request.identifier)
        notification = self.implementation._clear_notification_from_cache(identifier)

        if response.actionIdentifier == UNNotificationDefaultActionIdentifier:
            self.implementation.handle_clicked(identifier, notification)

        elif response.actionIdentifier == UNNotificationDismissActionIdentifier:
            self.implementation.handle_dismissed(identifier, notification)

        elif response.actionIdentifier == ReplyActionIdentifier:
            reply_text = py_from_ns(response.userText)
            self.implementation.handle_replied(identifier, reply_text, notification)

        else:
            action_id = py_from_ns(response.actionIdentifier)
            self.implementation.handle_button(identifier, action_id, notification)

        completion_handler()


class CocoaNotificationCenter(DesktopNotifierBackend):
    """UNUserNotificationCenter backend for macOS

    Can be used with macOS Catalina and newer. Both app name and bundle identifier
    will be ignored. The notification center automatically uses the values provided
    by the app bundle.

    :param app_name: The name of the app.
    """

    _to_native_urgency = {
        Urgency.Low: UNNotificationInterruptionLevel.Passive,
        Urgency.Normal: UNNotificationInterruptionLevel.Active,
        Urgency.Critical: UNNotificationInterruptionLevel.TimeSensitive,
    }

    def __init__(self, app_name: str, app_icon: Icon | None = None) -> None:
        super().__init__(app_name, app_icon)
        self.nc = UNUserNotificationCenter.currentNotificationCenter()
        self.nc_delegate = NotificationCenterDelegate.alloc().init()
        self.nc_delegate.implementation = self
        self.nc.delegate = self.nc_delegate

        self._clear_notification_categories()

    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send user notifications. If this is called for the
        first time for an app, the call will only return once the user has granted or
        denied the request. Otherwise, the call will just return the current
        authorisation status without prompting the user.

        :returns: Whether authorisation has been granted.
        """
        logger.debug("Requesting notification authorisation...")
        future: Future[tuple[bool, str]] = Future()

        def on_auth_completed(granted: bool, error: objc_id) -> None:
            ns_error = py_from_ns(error)
            future.set_result((granted, ns_error))

        self.nc.requestAuthorizationWithOptions(
            UNAuthorizationOptionAlert
            | UNAuthorizationOptionSound
            | UNAuthorizationOptionBadge,
            completionHandler=on_auth_completed,
        )

        has_authorization, error = await asyncio.wrap_future(future)

        if error:
            log_nserror(error, "Error requesting notification authorization")
        elif not has_authorization:
            logger.info("Not authorized to send notifications.")
        else:
            logger.debug("Authorized to send notifications")

        return has_authorization

    async def has_authorisation(self) -> bool:
        """Whether we have authorisation to send notifications."""
        future: Future[UNNotificationSettings] = Future()  # type:ignore[valid-type]

        def handler(settings: objc_id) -> None:
            settings = py_from_ns(settings)
            future.set_result(settings)
            future.set_result(settings)

        self.nc.getNotificationSettingsWithCompletionHandler(handler)

        settings = await asyncio.wrap_future(future)
        return settings.authorizationStatus in (  # type:ignore[attr-defined]
            UNAuthorizationStatusAuthorized,
            UNAuthorizationStatusProvisional,
            UNAuthorizationStatusEphemeral,
        )

    async def get_current_notifications(self) -> list[str]:
        future: Future[list[UNNotification]] = Future()  # type:ignore[valid-type]

        def handler(notifications: objc_id) -> None:
            notifications = py_from_ns(notifications)
            for notification in notifications:
                notification.retain()
            future.set_result(notifications)

        self.nc.getDeliveredNotificationsWithCompletionHandler(handler)

        notifications = await asyncio.wrap_future(future)
        identifiers = [
            str(n.request.identifier)  # type:ignore[attr-defined]
            for n in notifications
        ]

        for notification in notifications:
            notification.autorelease()  # type:ignore[attr-defined]

        return identifiers

    async def _send(self, notification: Notification) -> None:
        """
        Uses UNUserNotificationCenter to schedule a notification.

        :param notification: Notification to send.
        """
        # On macOS, we need to register a new notification category for every
        # unique set of buttons.
        category_id = await self._find_or_create_notification_category(notification)
        logger.debug("Notification category_id: %s", category_id)

        # Create the native notification and notification request.
        content = UNMutableNotificationContent.alloc().init()
        content.title = notification.title
        content.body = notification.message
        content.categoryIdentifier = category_id
        content.threadIdentifier = notification.thread
        if macos_version >= Version("12.0"):
            content.interruptionLevel = self._to_native_urgency[notification.urgency]

        if notification.sound:
            if notification.sound == DEFAULT_SOUND:
                content.sound = UNNotificationSound.defaultSound
            elif notification.sound.name:
                content.sound = UNNotificationSound.soundNamed(notification.sound.name)

        if notification.attachment:
            # Copy attachment to temporary file to ensure that it exists and that we can
            # access it. Invalid file paths can otherwise cause a segfault when creating
            # UNNotificationAttachment. The temporary file will be deleted by macOS
            # after usage.
            attachment_path = notification.attachment.as_path()
            tmp_dir = tempfile.mkdtemp()
            try:
                tmp_path = Path(tmp_dir) / attachment_path.name
                shutil.copy(attachment_path, tmp_path)
            except OSError:
                logger.warning("Could not access attachment file", exc_info=True)
            else:
                url = NSURL.fileURLWithPath(str(tmp_path), isDirectory=False)
                attachment = UNNotificationAttachment.attachmentWithIdentifier(
                    "", URL=url, options={}, error=None
                )
                content.attachments = [attachment]

        notification_request = UNNotificationRequest.requestWithIdentifier(
            notification.identifier, content=content, trigger=None
        )

        future: Future[NSError] = Future()  # type:ignore[valid-type]

        def handler(error: objc_id) -> None:
            ns_error = py_from_ns(error)
            future.set_result(ns_error)

        # Post the notification.
        self.nc.addNotificationRequest(
            notification_request, withCompletionHandler=handler
        )

        # Error handling.
        error = await asyncio.wrap_future(future)

        if error:
            log_nserror(error, "Error when scheduling notification")

    async def _find_or_create_notification_category(
        self, notification: Notification
    ) -> str:
        """
        Registers a new UNNotificationCategory for the given notification or retrieves
        an existing one.

        A new category is registered for each set of unique button titles, reply field
        title and reply field button title since on Apple platforms all of these
        elements are tied to a UNNotificationCategory.

        :param notification: Notification instance.
        :returns: The identifier of the existing or created notification category.
        """
        id_list = ["desktop-notifier"]
        for button in notification.buttons:
            id_list += [f"button-title-{button.title}"]

        if notification.reply_field:
            id_list += [f"reply-title-{notification.reply_field.title}"]
            id_list += [f"reply-button-title-{notification.reply_field.button_title}"]

        category_id = "_".join(id_list)

        # Retrieve existing categories. We do not cache this value because it may be
        # modified by other Python processes using desktop-notifier.
        categories = await self._get_notification_categories()
        category_ids = set(
            py_from_ns(c.identifier)
            for c in categories.allObjects()  # type:ignore[attr-defined]
        )

        # Register new category if necessary.
        if category_id not in category_ids:
            # Create action for each button.
            logger.debug("Creating new notification category: '%s'", category_id)
            actions = []

            if notification.reply_field:
                action = UNTextInputNotificationAction.actionWithIdentifier(
                    ReplyActionIdentifier,
                    title=notification.reply_field.title,
                    options=UNNotificationActionOptionNone,
                    textInputButtonTitle=notification.reply_field.button_title,
                    textInputPlaceholder="",
                )
                actions.append(action)

            for button in notification.buttons:
                action = UNNotificationAction.actionWithIdentifier(
                    button.identifier,
                    title=button.title,
                    options=UNNotificationActionOptionNone,
                )
                actions.append(action)

            # Add category for new set of buttons.
            new_categories = categories.setByAddingObject(  # type:ignore[attr-defined]
                UNNotificationCategory.categoryWithIdentifier(
                    category_id,
                    actions=actions,
                    intentIdentifiers=[],
                    options=UNNotificationCategoryOptionCustomDismissAction,
                )
            )
            self.nc.setNotificationCategories(new_categories)

        return category_id

    async def _get_notification_categories(self) -> NSSet:  # type:ignore[valid-type]
        """Returns the registered notification categories for this app / Python."""
        future: Future[NSSet] = Future()  # type:ignore[valid-type]

        def handler(categories: objc_id) -> None:
            categories = py_from_ns(categories)
            future.set_result(categories)

        self.nc.getNotificationCategoriesWithCompletionHandler(handler)

        return await asyncio.wrap_future(future)

    def _clear_notification_categories(self) -> None:
        """Clears all registered notification categories for this application."""
        empty_set = NSSet.alloc().init()
        self.nc.setNotificationCategories(empty_set)

    async def _clear(self, identifier: str) -> None:
        """
        Removes a notifications from the notification center

        :param identifier: Notification identifier.
        """
        self.nc.removeDeliveredNotificationsWithIdentifiers([identifier])

    async def _clear_all(self) -> None:
        """
        Clears all notifications from notification center. This method does not affect
        any notification requests that are scheduled, but have not yet been delivered.
        """
        self.nc.removeAllDeliveredNotifications()

    async def _get_capabilities(self) -> frozenset[Capability]:
        capabilities = {
            Capability.TITLE,
            Capability.MESSAGE,
            Capability.BUTTONS,
            Capability.REPLY_FIELD,
            Capability.ON_DISPATCHED,
            Capability.ON_CLICKED,
            Capability.ON_DISMISSED,
            Capability.SOUND,
            Capability.SOUND_NAME,
            Capability.THREAD,
            Capability.ATTACHMENT,
        }
        if macos_version >= Version("12.0"):
            capabilities.add(Capability.URGENCY)

        return frozenset(capabilities)


def log_nserror(error: NSError, prefix: str) -> None:  # type:ignore[valid-type]
    domain = str(error.domain)  # type:ignore[attr-defined]
    code = int(error.code)  # type:ignore[attr-defined]
    description = str(error.localizedDescription)  # type:ignore[attr-defined]

    logger.warning(
        "%s: domain=%s, code=%s, description=%s", prefix, domain, code, description
    )



# -*- coding: utf-8 -*-
"""
Desktop notifications for Windows, Linux, macOS, iOS and iPadOS.
"""
from .main import (
    DEFAULT_ICON,
    DEFAULT_SOUND,
    Attachment,
    Button,
    Capability,
    DesktopNotifier,
    Icon,
    Notification,
    ReplyField,
    Sound,
    Urgency,
)
from .sync import DesktopNotifierSync

__version__ = "6.1.1"
__author__ = "Sam Schott"
__url__ = "https://github.com/samschott/desktop-notifier"

__all__ = [
    "__version__",
    "__author__",
    "__url__",
    "Notification",
    "Button",
    "ReplyField",
    "Urgency",
    "Icon",
    "Sound",
    "Attachment",
    "DesktopNotifier",
    "DesktopNotifierSync",
    "Capability",
    "DEFAULT_SOUND",
    "DEFAULT_ICON",
]



"""
This module defines base classes for desktop notifications.
"""

from __future__ import annotations

import dataclasses
import logging
import uuid
from dataclasses import dataclass, field
from enum import Enum, auto
from importlib.resources import as_file, files
from pathlib import Path
from typing import Any, Callable
from urllib.parse import unquote, urlparse

__all__ = [
    "Capability",
    "FileResource",
    "Resource",
    "Icon",
    "Sound",
    "Attachment",
    "Button",
    "ReplyField",
    "Urgency",
    "AuthorisationError",
    "Notification",
    "DEFAULT_ICON",
    "DEFAULT_SOUND",
]


logger = logging.getLogger(__name__)
python_icon_path = as_file(
    files("desktop_notifier.resources") / "python.png"
).__enter__()


def uuid_str() -> str:
    return str(uuid.uuid4())


@dataclass(frozen=True)
class FileResource:
    """
    A file resource represented by a URI or path

    Only one of :attr:`path` or :attr:`uri` can be set.
    """

    path: Path | None = None
    """Path to a local file"""

    uri: str | None = None
    """URI reference to a file"""

    def __post_init__(self) -> None:
        fields = dataclasses.fields(self)
        set_fields = [f for f in fields if getattr(self, f.name) != f.default]
        if len(set_fields) > 1:
            raise RuntimeError("Only a single field can be set")
        if len(set_fields) == 0:
            field_names = [f.name for f in fields]
            raise RuntimeError(f"Either of {field_names} must be set")

    def as_uri(self) -> str:
        """
        Returns the represented resource as a URI string
        """
        if self.uri is not None:
            return self.uri
        if self.path is not None:
            return self.path.as_uri()
        raise AttributeError("No path or URI provided")

    def as_path(self) -> Path:
        """
        Returns the represented resource as a Path

        Note that any information about the URI scheme is lost on conversion.
        """
        if self.path is not None:
            return self.path
        if self.uri is not None:
            parsed_uri = urlparse(self.uri)
            return Path(unquote(parsed_uri.path))

        raise AttributeError("No path or URI provided")


@dataclass(frozen=True)
class Resource(FileResource):
    """
    A resource represented by a resource name, URI or path

    Only one of :attr:`path`, :attr:`uri` or :attr:`name` can be set.
    """

    name: str | None = None
    """Name of the system resource"""

    def as_name(self) -> str:
        if self.name is not None:
            return self.name
        raise AttributeError("No resource name provided")

    def is_named(self) -> bool:
        """Returns whether the instance was initialized with ``name``"""
        return self.name is not None

    def is_file(self) -> bool:
        """Returns whether the instance was initialized with ``path`` or ``uri``"""
        return self.path is not None or self.uri is not None


@dataclass(frozen=True)
class Icon(Resource):
    """An icon represented by an icon name, URI or path"""

    pass


@dataclass(frozen=True)
class Attachment(FileResource):
    """An attachment represented by a URI or path"""

    pass


@dataclass(frozen=True)
class Sound(Resource):
    """A sound represented by a sound name, URI or path"""

    pass


DEFAULT_ICON: Icon = Icon(path=python_icon_path)
"""Python icon"""

DEFAULT_SOUND: Sound = Sound(name="default")
"""Default system notification sound"""


class AuthorisationError(Exception):
    """Raised when we are not authorised to send notifications"""


class Urgency(Enum):
    """Enumeration of notification levels

    The interpretation and visuals depend on the platform.
    """

    Critical = "critical"
    """For critical errors."""

    Normal = "normal"
    """Default platform notification level."""

    Low = "low"
    """Low priority notification."""


@dataclass(frozen=True)
class Button:
    """A button for interactive notifications"""

    title: str
    """The localized button title"""

    on_pressed: Callable[[], Any] | None = None
    """Method to call when the button is pressed"""

    identifier: str = dataclasses.field(default_factory=uuid_str)
    """A unique identifier to use in callbacks to specify with button was clicked"""


@dataclass(frozen=True)
class ReplyField:
    """A text field for interactive notifications"""

    title: str = "Reply"
    """A title for the field itself. On macOS, this will be the title of a button to
    show the field."""

    button_title: str = "Send"
    """The title of the button to send the reply"""

    on_replied: Callable[[str], Any] | None = None
    """Method to call when the 'reply' button is pressed"""


@dataclass(frozen=True)
class Notification:
    """A desktop notification

    Some properties of a notification may be ignored or interpreted differently
    depending on the platform.

    Callbacks for interactions will be executed on the Python process that scheduled the
    notification and only as long as the DesktopNotifier instance that scheduled the
    notification still exists.

    Install handlers on the DesktopNotifier instance itself to respond to interactions
    with notification from your app while it was not running.
    """

    title: str
    """Notification title"""

    message: str
    """Notification message"""

    urgency: Urgency = Urgency.Normal
    """Notification urgency. Can determine stickiness, notification appearance and
    break through silencing."""

    icon: Icon | None = None
    """Icon to use for the notification"""

    buttons: tuple[Button, ...] = field(default_factory=tuple)
    """Buttons shown on an interactive notification"""

    reply_field: ReplyField | None = None
    """Text field shown on an interactive notification. This can be used for example
    for messaging apps to reply directly from the notification."""

    on_dispatched: Callable[[], Any] | None = None
    """Method to call when the notification was sent to the notifications server for display"""

    on_clicked: Callable[[], Any] | None = None
    """Method to call when the notification is clicked"""

    on_dismissed: Callable[[], Any] | None = None
    """Method to call when the notification is dismissed"""

    attachment: Attachment | None = None
    """A file attached to the notification which may be displayed as a preview"""

    sound: Sound | None = None
    """A sound to play on notification"""

    thread: str | None = None
    """An identifier to group related notifications together, e.g., from a chat space"""

    timeout: int = -1
    """Duration in seconds for which the notification is shown"""

    identifier: str = field(default_factory=uuid_str)
    """A unique identifier for this notification. Generated automatically if not
    passed by the client."""

    _buttons_dict: dict[str, Button] = field(default_factory=dict)

    def __post_init__(self) -> None:
        for button in self.buttons:
            self._buttons_dict[button.identifier] = button

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}(identifier='{self.identifier}', "
            f"title='{self.title}', message='{self.message}')>"
        )


class Capability(Enum):
    """Notification capabilities that can be supported by a platform"""

    APP_NAME = auto()
    """Supports setting a custom app name"""

    TITLE = auto()
    """Supports setting a notification title"""

    MESSAGE = auto()
    """Supports setting a notification message"""

    URGENCY = auto()
    """Supports different urgency levels"""

    ICON = auto()
    """Supports custom notification icons"""

    ICON_FILE = auto()
    """Supports setting a custom icon from a user-provided file"""

    ICON_NAME = auto()
    """Supports setting a named system icon as notification icon"""

    BUTTONS = auto()
    """Supports at least two notification buttons"""

    REPLY_FIELD = auto()
    """Supports reply fields"""

    ATTACHMENT = auto()
    """Supports notification attachments. Allowed file types vary by platform."""

    ON_DISPATCHED = auto()
    """Supports on-dispatched callbacks"""

    ON_CLICKED = auto()
    """Supports on-clicked callbacks"""

    ON_DISMISSED = auto()
    """Supports on-dismissed callbacks"""

    SOUND = auto()
    """Supports custom notification sounds"""

    SOUND_FILE = auto()
    """Supports setting a custom sound from a user-provided file"""

    SOUND_NAME = auto()
    """Supports setting a named system sound as notification sound"""

    THREAD = auto()
    """Supports grouping notifications by topic thread"""

    TIMEOUT = auto()
    """Supports notification timeouts"""



import time

from desktop_notifier import (
    DEFAULT_SOUND,
    Button,
    DesktopNotifierSync,
    ReplyField,
    Urgency,
)


def wait_for_notifications(
    notifier: DesktopNotifierSync, notification_count: int = 1, timeout_sec: float = 2.0
) -> None:
    t0 = time.monotonic()

    while time.monotonic() - t0 < timeout_sec:
        if len(notifier.get_current_notifications()) == notification_count:
            return
        time.sleep(0.2)

    raise TimeoutError("Timed out while waiting for notifications")


def test_send(notifier_sync: DesktopNotifierSync) -> None:
    notification = notifier_sync.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
        urgency=Urgency.Critical,
        buttons=[
            Button(
                title="Mark as read",
                on_pressed=lambda: print("Marked as read"),
            )
        ],
        reply_field=ReplyField(
            title="Reply",
            button_title="Send",
            on_replied=lambda text: print("Brutus replied:", text),
        ),
        on_clicked=lambda: print("Notification clicked"),
        on_dismissed=lambda: print("Notification dismissed"),
        sound=DEFAULT_SOUND,
        thread="test_notifications",
        timeout=5,
    )
    wait_for_notifications(notifier_sync)
    assert notification in notifier_sync.get_current_notifications()


def test_clear(notifier_sync: DesktopNotifierSync) -> None:
    n0 = notifier_sync.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    n1 = notifier_sync.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    wait_for_notifications(notifier_sync, 2)
    current_notifications = notifier_sync.get_current_notifications()

    assert n0 in current_notifications
    assert n1 in current_notifications

    notifier_sync.clear(n0)
    assert n0 not in notifier_sync.get_current_notifications()


def test_clear_all(notifier_sync: DesktopNotifierSync) -> None:
    n0 = notifier_sync.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    n1 = notifier_sync.send(
        title="Julius Caesar",
        message="Et tu, Brute?",
    )
    wait_for_notifications(notifier_sync, 2)
    current_notifications = notifier_sync.get_current_notifications()

    assert n0 in current_notifications
    assert n1 in current_notifications

    notifier_sync.clear_all()
    assert len(notifier_sync.get_current_notifications()) == 0



# -*- coding: utf-8 -*-
"""
Notification backend for Linux

Includes an implementation to send desktop notifications over Dbus. Responding to user
interaction with a notification requires a running asyncio event loop.
"""
from __future__ import annotations

import logging
from typing import TypeVar

from bidict import bidict
from dbus_fast.aio.message_bus import MessageBus
from dbus_fast.aio.proxy_object import ProxyInterface
from dbus_fast.errors import DBusError
from dbus_fast.signature import Variant

from ..common import Capability, Icon, Notification, Urgency
from .base import DesktopNotifierBackend

__all__ = ["DBusDesktopNotifier"]

logger = logging.getLogger(__name__)

T = TypeVar("T")

NOTIFICATION_CLOSED_EXPIRED = 1
NOTIFICATION_CLOSED_DISMISSED = 2
NOTIFICATION_CLOSED_PROGRAMMATICALLY = 3
NOTIFICATION_CLOSED_UNDEFINED = 4

DEFAULT_ACTION_KEY = "default"


class DBusDesktopNotifier(DesktopNotifierBackend):
    """DBus notification backend for Linux

    This implements the org.freedesktop.Notifications standard. The DBUS connection is
    created in a thread with a running asyncio loop to handle clicked notifications.

    :param app_name: The name of the app.
    """

    to_native_urgency = {
        Urgency.Low: Variant("y", 0),
        Urgency.Normal: Variant("y", 1),
        Urgency.Critical: Variant("y", 2),
    }

    supported_hint_signatures = {"a{sv}", "a{ss}"}

    def __init__(self, app_name: str, app_icon: Icon | None = None) -> None:
        super().__init__(app_name, app_icon)
        self.interface: ProxyInterface | None = None
        self._platform_to_interface_notification_identifier: bidict[int, str] = bidict()

    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send notifications.

        :returns: Whether authorisation has been granted.
        """
        return True

    async def has_authorisation(self) -> bool:
        """
        Whether we have authorisation to send notifications.
        """
        return True

    async def _init_dbus(self) -> ProxyInterface:
        self.bus = await MessageBus().connect()
        introspection = await self.bus.introspect(
            "org.freedesktop.Notifications", "/org/freedesktop/Notifications"
        )
        self.proxy_object = self.bus.get_proxy_object(
            "org.freedesktop.Notifications",
            "/org/freedesktop/Notifications",
            introspection,
        )
        self.interface = self.proxy_object.get_interface(
            "org.freedesktop.Notifications"
        )

        # Some older interfaces may not support notification actions.
        if hasattr(self.interface, "on_notification_closed"):
            self.interface.on_notification_closed(self._on_closed)

        if hasattr(self.interface, "on_action_invoked"):
            self.interface.on_action_invoked(self._on_action)

        return self.interface

    async def _send(self, notification: Notification) -> None:
        """
        Asynchronously sends a notification via the Dbus interface.

        :param notification: Notification to send.
        """
        if not self.interface:
            self.interface = await self._init_dbus()

        actions = []

        # The "default" action is invoked when clicking on the notification body.
        if Capability.ON_CLICKED in await self.get_capabilities():
            actions += [DEFAULT_ACTION_KEY, ""]

        for button in notification.buttons:
            actions += [button.identifier, button.title]

        hints_v: dict[str, Variant] = dict()
        hints_v["urgency"] = self.to_native_urgency[notification.urgency]

        if notification.sound:
            if notification.sound.is_named():
                hints_v["sound-name"] = Variant("s", notification.sound.name)
            else:
                hints_v["sound-file"] = Variant("s", notification.sound.as_uri())

        if notification.attachment:
            hints_v["image-path"] = Variant("s", notification.attachment.as_uri())

        # The current notification spec defines hints as a Dbus dictionary type 'a{sv}',
        # represented in Python as dict[str, Variant]. However, some older notification
        # servers expect 'a{ss}' (Python dict[str, str]). We therefore check the
        # expected argument type at runtime and cast arguments accordingly.
        # See https://github.com/samschott/desktop-notifier/issues/143.
        hints_signature = get_hints_signature(self.interface)

        if hints_signature == "":
            logger.warning("Notification server not supported")
            return

        hints: dict[str, str] | dict[str, Variant]

        if hints_signature == "a{sv}":
            hints = hints_v
        elif hints_signature == "a{ss}":
            hints = {k: str(v.value) for k, v in hints_v.items()}
        else:
            hints = {}

        timeout = notification.timeout * 1000 if notification.timeout != -1 else -1

        icon: str = ""
        if notification.icon:
            icon = (
                notification.icon.as_name()
                if notification.icon.is_named()
                else notification.icon.as_uri()
            )
        elif self.app_icon:
            icon = (
                self.app_icon.as_name()
                if self.app_icon.is_named()
                else self.app_icon.as_uri()
            )

        # dbus_next proxy APIs are generated at runtime. Silence the type checker but
        # raise an AttributeError if required.
        platform_id = await self.interface.call_notify(  # type:ignore[attr-defined]
            self.app_name,
            0,
            icon,
            notification.title,
            notification.message,
            actions,
            hints,
            timeout,
        )
        self._platform_to_interface_notification_identifier[platform_id] = (
            notification.identifier
        )

    async def _clear(self, identifier: str) -> None:
        """
        Asynchronously removes a notification from the notification center
        """
        if not self.interface:
            self.interface = await self._init_dbus()

        platform_id = self._platform_to_interface_notification_identifier.inverse[
            identifier
        ]

        try:
            # dbus_next proxy APIs are generated at runtime. Silence the type checker
            # but raise an AttributeError if required.
            await self.interface.call_close_notification(  # type:ignore[attr-defined]
                platform_id
            )
        except DBusError:
            # Notification was already closed.
            # See https://specifications.freedesktop.org/notification-spec/latest/protocol.html#command-close-notification
            pass

        try:
            del self._platform_to_interface_notification_identifier.inverse[identifier]
        except KeyError:
            # Popping may have been handled already by _on_close callback.
            pass

    async def _clear_all(self) -> None:
        """
        Asynchronously clears all notifications from notification center
        """
        for identifier in list(
            self._platform_to_interface_notification_identifier.values()
        ):
            await self._clear(identifier)

    # Note that _on_action and _on_closed might be called for the same notification
    # with some notification servers. This is not a problem because the _on_action
    # call will come first, in which case we are no longer interested in calling the
    # _on_closed callback.

    def _on_action(self, nid: int, action_key: str) -> None:
        """
        Called when the user performs a notification action. This will invoke the
        handler callback.

        :param nid: The platform's notification ID as an integer.
        :param action_key: A string identifying the action to take. We choose those keys
            ourselves when scheduling the notification.
        """
        identifier = self._platform_to_interface_notification_identifier.pop(nid, "")
        notification = self._clear_notification_from_cache(identifier)

        if not notification:
            return

        if action_key == DEFAULT_ACTION_KEY:
            self.handle_clicked(identifier, notification)
            return

        self.handle_button(identifier, action_key, notification)

    def _on_closed(self, nid: int, reason: int) -> None:
        """
        Called when the user closes a notification. This will invoke the registered
        callback.

        :param nid: The platform's notification ID as an integer.
        :param reason: An integer describing the reason why the notification was closed.
        """
        identifier = self._platform_to_interface_notification_identifier.pop(nid, "")
        notification = self._clear_notification_from_cache(identifier)

        if not notification:
            return

        if reason == NOTIFICATION_CLOSED_DISMISSED:
            self.handle_dismissed(identifier, notification)

    async def _get_capabilities(self) -> frozenset[Capability]:
        if not self.interface:
            self.interface = await self._init_dbus()

        capabilities = {
            Capability.APP_NAME,
            Capability.ICON,
            Capability.TITLE,
            Capability.TIMEOUT,
            Capability.URGENCY,
            Capability.ON_DISPATCHED,
        }

        # Capabilities supported by some notification servers.
        # See https://specifications.freedesktop.org/notification-spec/notification-spec-latest.html#protocol.
        if hasattr(self.interface, "on_notification_closed"):
            capabilities.add(Capability.ON_DISMISSED)

        server_info = (
            await self.interface.call_get_server_information()  # type:ignore[attr-defined]
        )

        # xfce4-notifyd does not support a "default" action when the notification is
        # clicked. See https://docs.xfce.org/apps/xfce4-notifyd/spec.
        if server_info[0] != "Xfce Notify Daemon":
            capabilities.add(Capability.ON_CLICKED)

        cps = await self.interface.call_get_capabilities()  # type:ignore[attr-defined]

        if "actions" in cps:
            capabilities.add(Capability.BUTTONS)
        if "body" in cps:
            capabilities.add(Capability.MESSAGE)
        if "sound" in cps:
            capabilities.add(Capability.SOUND)
            capabilities.add(Capability.SOUND_NAME)

        hints_signature = get_hints_signature(self.interface)
        if hints_signature not in self.supported_hint_signatures:
            # Any hint-based capabilities are not supported because we got an unexpected
            # DBus interface.
            capabilities.discard(Capability.SOUND)
            capabilities.discard(Capability.SOUND_NAME)
            capabilities.discard(Capability.URGENCY)

        return frozenset(capabilities)


def get_hints_signature(interface: ProxyInterface) -> str:
    """Returns the dbus type signature for the hints argument"""
    methods = interface.introspection.methods
    notify_method = next(m for m in methods if m.name == "Notify")
    try:
        return str(notify_method.in_args[6].signature)
    except IndexError:
        return ""



# -*- coding: utf-8 -*-
"""
This module defines the abstract implementation class that backends must inherit from.
"""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Callable

from ..common import Capability, Icon, Notification

__all__ = [
    "DesktopNotifierBackend",
]


logger = logging.getLogger(__name__)


class DesktopNotifierBackend(ABC):
    """Base class for desktop notifier implementations

    :param app_name: Name to identify the application in the notification center.
    """

    def __init__(self, app_name: str, app_icon: Icon | None = None) -> None:
        self.app_name = app_name
        self.app_icon = app_icon
        self._notification_cache: dict[str, Notification] = dict()

        self.on_dispatched: Callable[[str], Any] | None = None
        self.on_clicked: Callable[[str], Any] | None = None
        self.on_dismissed: Callable[[str], Any] | None = None
        self.on_button_pressed: Callable[[str, str], Any] | None = None
        self.on_replied: Callable[[str, str], Any] | None = None

        self._capabilities: frozenset[Capability] | None = None

    @abstractmethod
    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send notifications.

        :returns: Whether authorisation has been granted.
        """
        ...

    @abstractmethod
    async def has_authorisation(self) -> bool:
        """
        Returns whether we have authorisation to send notifications.
        """
        ...

    async def send(self, notification: Notification) -> None:
        """
        Sends a desktop notification.

        :param notification: Notification to send.
        """
        try:
            await self._send(notification)
        except Exception:
            # Notifications can fail for many reasons:
            # The dbus service may not be available, we might be in a headless session,
            # etc. Since notifications are not critical to an application, we only emit
            # a warning.
            logger.warning("Notification failed", exc_info=True)
        else:
            logger.debug("Notification sent: %s", notification)
            self._notification_cache[notification.identifier] = notification

            self.handle_dispatched(notification.identifier, notification)

    def _clear_notification_from_cache(self, identifier: str) -> Notification | None:
        """
        Removes the notification from our cache. Should be called by backends when the
        notification is closed.
        """
        return self._notification_cache.pop(identifier, None)

    @abstractmethod
    async def _send(self, notification: Notification) -> None:
        """
        Method to send a notification via the platform.

        Implementations must raise an exception when the notification could not be
        delivered. If the notification could be delivered but not fully as intended,
        e.g., because associated resources could not be loaded, implementations should
        emit a log message of level warning.

        :param notification: Notification to send.
        """
        ...

    async def get_current_notifications(self) -> list[str]:
        """Returns identifiers of all currently displayed notifications for this app."""
        return list(self._notification_cache.keys())

    async def clear(self, identifier: str) -> None:
        """
        Removes the given notification from the notification center. This is a wrapper
        method which mostly performs housekeeping of notifications ID and calls
        :meth:`_clear` to actually clear the notification. Platform implementations
        must implement :meth:`_clear`.

        :param identifier: Notification identifier.
        """
        await self._clear(identifier)
        self._clear_notification_from_cache(identifier)

    @abstractmethod
    async def _clear(self, identifier: str) -> None:
        """
        Removes the given notification from the notification center.

        :param identifier: Notification identifier.
        """
        ...

    async def clear_all(self) -> None:
        """
        Clears all notifications from the notification center. This is a wrapper method
        which mostly performs housekeeping of notifications ID and calls
        :meth:`_clear_all` to actually clear the notifications. Platform implementations
        must implement :meth:`_clear_all`.
        """

        await self._clear_all()
        self._notification_cache.clear()

    @abstractmethod
    async def _clear_all(self) -> None:
        """
        Clears all notifications from the notification center.
        """
        ...

    @abstractmethod
    async def _get_capabilities(self) -> frozenset[Capability]:
        """
        Returns the functionality supported by the notification server.
        """
        ...

    async def get_capabilities(self) -> frozenset[Capability]:
        """
        Returns the functionality supported by the notification server. Caches the
        result.
        """
        if not self._capabilities:
            self._capabilities = await self._get_capabilities()
        return self._capabilities

    def handle_dispatched(
        self, identifier: str, notification: Notification | None = None
    ) -> None:
        if notification and notification.on_dispatched:
            notification.on_dispatched()
        elif self.on_dispatched:
            self.on_dispatched(identifier)

    def handle_clicked(
        self, identifier: str, notification: Notification | None = None
    ) -> None:
        if notification and notification.on_clicked:
            notification.on_clicked()
        elif self.on_clicked:
            self.on_clicked(identifier)

    def handle_dismissed(
        self, identifier: str, notification: Notification | None = None
    ) -> None:
        if notification and notification.on_dismissed:
            notification.on_dismissed()
        elif self.on_dismissed:
            self.on_dismissed(identifier)

    def handle_replied(
        self, identifier: str, reply_text: str, notification: Notification | None = None
    ) -> None:
        if (
            notification
            and notification.reply_field
            and notification.reply_field.on_replied
        ):
            notification.reply_field.on_replied(reply_text)
        elif self.on_replied:
            self.on_replied(identifier, reply_text)

    def handle_button(
        self,
        identifier: str,
        button_identifier: str,
        notification: Notification | None = None,
    ) -> None:
        if notification and button_identifier in notification._buttons_dict:
            button = notification._buttons_dict[button_identifier]
        else:
            button = None

        if button and button.on_pressed:
            button.on_pressed()
        elif self.on_button_pressed:
            self.on_button_pressed(identifier, button_identifier)



import platform
from unittest.mock import Mock

import pytest

from desktop_notifier import (
    Button,
    Capability,
    DesktopNotifier,
    Notification,
    ReplyField,
)

from .backends import (
    simulate_button_pressed,
    simulate_clicked,
    simulate_dismissed,
    simulate_replied,
)

if platform.system() == "Windows":
    pytest.skip("Windows test infra missing", allow_module_level=True)


async def check_supported(notifier: DesktopNotifier, capability: Capability) -> None:
    capabilities = await notifier.get_capabilities()
    if capability not in capabilities:
        pytest.skip(f"{capability} not supported by {notifier} backend")


@pytest.mark.asyncio
async def test_dispatched_callback_called(notifier: DesktopNotifier) -> None:
    class_handler = Mock()
    notification_handler = Mock()
    notifier.on_dispatched = class_handler
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        on_dispatched=notification_handler,
    )

    await notifier.send_notification(notification)

    class_handler.assert_not_called()
    notification_handler.assert_called_once()


@pytest.mark.asyncio
async def test_clicked_callback_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.ON_CLICKED)

    class_handler = Mock()
    notification_handler = Mock()
    notifier.on_clicked = class_handler
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        on_clicked=notification_handler,
    )

    identifier = await notifier.send_notification(notification)
    simulate_clicked(notifier, identifier)

    class_handler.assert_not_called()
    notification_handler.assert_called_once()


@pytest.mark.asyncio
async def test_clicked_callback_dismissed_not_called(notifier: DesktopNotifier) -> None:
    """
    Dbus may send an on_dismissed event any time a notification is closed. Ensure that
    its callback is not triggered on different types of interactions.
    """
    await check_supported(notifier, Capability.ON_CLICKED)

    on_clicked = Mock()
    on_dismissed = Mock()
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        on_clicked=on_clicked,
        on_dismissed=on_dismissed,
    )

    identifier = await notifier.send_notification(notification)
    simulate_clicked(notifier, identifier)

    on_dismissed.assert_not_called()
    on_clicked.assert_called_once()


@pytest.mark.asyncio
async def test_dismissed_callback_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.ON_DISMISSED)

    class_handler = Mock()
    notification_handler = Mock()
    notifier.on_dismissed = class_handler
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        on_dismissed=notification_handler,
    )

    identifier = await notifier.send_notification(notification)
    simulate_dismissed(notifier, identifier)

    class_handler.assert_not_called()
    notification_handler.assert_called_once()


@pytest.mark.asyncio
async def test_button_pressed_callback_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.BUTTONS)

    class_handler = Mock()
    notification_b0_handler = Mock()
    notification_b1_handler = Mock()
    notifier.on_button_pressed = class_handler
    button0 = Button(title="Button 0", on_pressed=notification_b0_handler)
    button1 = Button(title="Button 1", on_pressed=notification_b1_handler)
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        buttons=(button0, button1),
    )

    identifier = await notifier.send_notification(notification)
    simulate_button_pressed(notifier, identifier, button1.identifier)

    class_handler.assert_not_called()
    notification_b1_handler.assert_called_once()


@pytest.mark.asyncio
async def test_replied_callback_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.REPLY_FIELD)
    class_handler = Mock()
    notification_handler = Mock()

    notifier.on_replied = class_handler
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        reply_field=ReplyField(on_replied=notification_handler),
    )

    identifier = await notifier.send_notification(notification)
    simulate_replied(notifier, identifier, "A notification response")

    class_handler.assert_not_called()
    notification_handler.assert_called_with("A notification response")


@pytest.mark.asyncio
async def test_dispatched_fallback_handler_called(notifier: DesktopNotifier) -> None:
    class_handler = Mock()
    notifier.on_dispatched = class_handler
    notification = Notification(title="Julius Caesar", message="Et tu, Brute?")

    identifier = await notifier.send_notification(notification)

    class_handler.assert_called_with(identifier)


@pytest.mark.asyncio
async def test_clicked_fallback_handler_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.ON_CLICKED)

    class_handler = Mock()
    notifier.on_clicked = class_handler
    notification = Notification(title="Julius Caesar", message="Et tu, Brute?")

    identifier = await notifier.send_notification(notification)
    simulate_clicked(notifier, identifier)

    class_handler.assert_called_with(identifier)


@pytest.mark.asyncio
async def test_dismissed_fallback_handler_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.ON_DISMISSED)

    class_handler = Mock()
    notifier.on_dismissed = class_handler
    notification = Notification(title="Julius Caesar", message="Et tu, Brute?")

    identifier = await notifier.send_notification(notification)
    simulate_dismissed(notifier, identifier)

    class_handler.assert_called_with(identifier)


@pytest.mark.asyncio
async def test_button_pressed_fallback_handler_called(
    notifier: DesktopNotifier,
) -> None:
    await check_supported(notifier, Capability.BUTTONS)

    class_handler = Mock()
    notifier.on_button_pressed = class_handler
    button0 = Button(title="Button 1")
    button1 = Button(title="Button 2")
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        buttons=(button0, button1),
    )

    identifier = await notifier.send_notification(notification)
    simulate_button_pressed(notifier, identifier, button1.identifier)

    class_handler.assert_called_once_with(identifier, button1.identifier)


@pytest.mark.asyncio
async def test_replied_fallback_handler_called(notifier: DesktopNotifier) -> None:
    await check_supported(notifier, Capability.REPLY_FIELD)

    class_handler = Mock()
    notifier.on_replied = class_handler
    notification = Notification(
        title="Julius Caesar",
        message="Et tu, Brute?",
        reply_field=ReplyField(),
    )

    identifier = await notifier.send_notification(notification)
    simulate_replied(notifier, identifier, "A notification response")

    class_handler.assert_called_with(identifier, "A notification response")



# -*- coding: utf-8 -*-
"""
Notification backend for Windows

Unlike other platforms, sending rich "toast" notifications cannot be done via FFI /
ctypes because the C winapi only supports basic notifications with a title and message.
This backend therefore requires interaction with the Windows Runtime and uses the winrt
package with compiled components.
"""
from __future__ import annotations

import logging
import sys
import winreg
import asyncio
from collections.abc import Callable
from typing import TypeVar
from typing_extensions import ParamSpec
from xml.etree.ElementTree import Element, SubElement, tostring

from winrt.system import Object as WinRTObject
from winrt.system import unbox_string
from winrt.windows.applicationmodel.core import CoreApplication
from winrt.windows.data.xml.dom import XmlDocument
from winrt.windows.ui.notifications import (
    NotificationSetting,
    ToastActivatedEventArgs,
    ToastDismissalReason,
    ToastDismissedEventArgs,
    ToastFailedEventArgs,
    ToastNotification,
    ToastNotificationManager,
    ToastNotificationPriority,
)

# local imports
from ..common import DEFAULT_SOUND, Capability, Icon, Notification, Urgency
from .base import DesktopNotifierBackend

__all__ = ["WinRTDesktopNotifier"]

logger = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")

DEFAULT_GROUP = "desktop-notifier-group"
DEFAULT_ACTION = "default"
REPLY_ACTION = "action=reply&amp"
BUTTON_ACTION_PREFIX = "action=button&amp;id="
REPLY_TEXTBOX_NAME = "textBox"


def register_hkey(app_id: str, app_name: str) -> None:
    # mypy type guard
    if not sys.platform == "win32":
        return

    winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)
    key_path = f"SOFTWARE\\Classes\\AppUserModelId\\{app_id}"
    with winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, key_path) as master_key:
        winreg.SetValueEx(master_key, "DisplayName", 0, winreg.REG_SZ, app_name)


class WinRTDesktopNotifier(DesktopNotifierBackend):
    """Notification backend for the Windows Runtime

    :param app_name: The name of the app.
    """

    _to_native_urgency = {
        Urgency.Low: ToastNotificationPriority.DEFAULT,
        Urgency.Normal: ToastNotificationPriority.DEFAULT,
        Urgency.Critical: ToastNotificationPriority.HIGH,
    }

    def __init__(self, app_name: str, app_icon: Icon | None = None) -> None:
        super().__init__(app_name, app_icon)

        manager = ToastNotificationManager.get_default()

        if not manager:
            raise RuntimeError("Could not get ToastNotificationManagerForUser")

        self.manager = manager

        # Prefer using the real App ID if detected, fall back to user-provided name
        # and icon otherwise.
        if CoreApplication.id != "":
            self.app_id = CoreApplication.id
        else:
            self.app_id = app_name
            register_hkey(app_id=app_name, app_name=app_name)

        notifier = self.manager.create_toast_notifier_with_id(self.app_id)

        if not notifier:
            raise RuntimeError(f"Could not get ToastNotifier for app_id: {self.app_id}")

        self.notifier = notifier

    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send notifications.

        :returns: Whether authorisation has been granted.
        """
        return await self.has_authorisation()

    async def has_authorisation(self) -> bool:
        """
        Whether we have authorisation to send notifications.
        """
        try:
            return bool(self.notifier.setting == NotificationSetting.ENABLED)
        except OSError:
            # See https://github.com/samschott/desktop-notifier/issues/95.
            return True

    async def _send(self, notification: Notification) -> None:
        """
        Asynchronously sends a notification.

        :param notification: Notification to send.
        """
        toast_xml = Element("toast", {"launch": DEFAULT_ACTION})
        visual_xml = SubElement(toast_xml, "visual")
        actions_xml = SubElement(toast_xml, "actions")

        if notification.thread:
            SubElement(
                toast_xml,
                "header",
                {
                    "id": notification.thread,
                    "title": notification.thread,
                    "arguments": DEFAULT_ACTION,
                    "activationType": "background",
                },
            )

        binding = SubElement(visual_xml, "binding", {"template": "ToastGeneric"})

        title_xml = SubElement(binding, "text")
        title_xml.text = notification.title

        message_xml = SubElement(binding, "text")
        message_xml.text = notification.message

        icon_obj: Icon | None = None
        if notification.icon:
            icon_obj = notification.icon
        elif self.app_icon:
            icon_obj = self.app_icon
        if icon_obj and icon_obj.is_file():
            SubElement(
                binding,
                "image",
                {
                    "placement": "appLogoOverride",
                    "src": icon_obj.as_uri(),
                },
            )

        if notification.attachment:
            SubElement(
                binding,
                "image",
                {"placement": "hero", "src": notification.attachment.as_uri()},
            )

        if notification.reply_field:
            SubElement(
                actions_xml,
                "input",
                {"id": REPLY_TEXTBOX_NAME, "type": "text"},
            )
            reply_button_xml = SubElement(
                actions_xml,
                "action",
                {
                    "content": notification.reply_field.button_title,
                    "activationType": "background",
                    "arguments": "action=reply&amp",
                },
            )

            # If there are no other buttons, show the reply button next to the text
            # field. Otherwise, show it above other buttons.
            if not notification.buttons:
                reply_button_xml.set("hint-inputId", REPLY_TEXTBOX_NAME)

        for button in notification.buttons:
            SubElement(
                actions_xml,
                "action",
                {
                    "content": button.title,
                    "activationType": "background",
                    "arguments": BUTTON_ACTION_PREFIX + button.identifier,
                },
            )

        if notification.sound:
            if notification.sound == DEFAULT_SOUND:
                sound_attr = {"src": "ms-winsoundevent:Notification.Default"}
            elif notification.sound.name:
                sound_attr = {"src": notification.sound.name}
            else:
                sound_attr = {"src": notification.sound.as_uri()}
        else:
            sound_attr = {"silent": "true"}

        SubElement(toast_xml, "audio", sound_attr)

        xml_document = XmlDocument()
        xml_document.load_xml(tostring(toast_xml, encoding="unicode"))

        native = ToastNotification(xml_document)
        native.tag = notification.identifier
        native.group = DEFAULT_GROUP
        native.priority = self._to_native_urgency[notification.urgency]

        native.add_activated(eventloop_wrapper(self._on_activated))
        native.add_dismissed(eventloop_wrapper(self._on_dismissed))
        native.add_failed(eventloop_wrapper(self._on_failed))

        self.notifier.show(native)

    def _on_activated(
        self, sender: ToastNotification | None, boxed_activated_args: WinRTObject | None
    ) -> None:
        if not sender:
            return

        notification = self._clear_notification_from_cache(sender.tag)

        if not boxed_activated_args:
            return

        activated_args = boxed_activated_args.as_(ToastActivatedEventArgs)
        action_id = activated_args.arguments

        if action_id == DEFAULT_ACTION:
            self.handle_clicked(sender.tag, notification)

        elif action_id == REPLY_ACTION and activated_args.user_input:
            boxed_reply = activated_args.user_input[REPLY_TEXTBOX_NAME]
            reply = unbox_string(boxed_reply)
            self.handle_replied(sender.tag, reply, notification)

        elif action_id.startswith(BUTTON_ACTION_PREFIX):
            button_id = action_id.replace(BUTTON_ACTION_PREFIX, "")
            self.handle_button(sender.tag, button_id, notification)

    def _on_dismissed(
        self,
        sender: ToastNotification | None,
        dismissed_args: ToastDismissedEventArgs | None,
    ) -> None:
        if not sender:
            return

        notification = self._clear_notification_from_cache(sender.tag)

        if (
            dismissed_args
            and dismissed_args.reason == ToastDismissalReason.USER_CANCELED
        ):
            self.handle_dismissed(sender.tag, notification)

    def _on_failed(
        self, sender: ToastNotification | None, failed_args: ToastFailedEventArgs | None
    ) -> None:
        if not sender:
            return

        self._clear_notification_from_cache(sender.tag)
        if failed_args:
            logger.warning(
                "Notification '%s' failed with error code %s",
                sender.tag,
                failed_args.error_code.value,
            )
        else:
            logger.warning("Notification '%s' failed with unknown error", sender.tag)

    async def _clear(self, identifier: str) -> None:
        """
        Asynchronously removes a notification from the notification center.
        """
        if self.manager.history:
            self.manager.history.remove_grouped_tag_with_id(
                identifier, DEFAULT_GROUP, self.app_id
            )

    async def _clear_all(self) -> None:
        """
        Asynchronously clears all notifications from notification center.
        """
        if self.manager.history:
            self.manager.history.clear_with_id(self.app_id)

    async def get_current_notifications(self) -> list[str]:
        if self.manager.history:
            notifications = self.manager.history.get_history_with_id(self.app_id)
            return [n.tag for n in notifications]
        return await super().get_current_notifications()

    async def _get_capabilities(self) -> frozenset[Capability]:
        capabilities = {
            Capability.TITLE,
            Capability.MESSAGE,
            Capability.ICON,
            Capability.BUTTONS,
            Capability.REPLY_FIELD,
            Capability.ON_DISPATCHED,
            Capability.ON_CLICKED,
            Capability.ON_DISMISSED,
            Capability.THREAD,
            Capability.ATTACHMENT,
            Capability.SOUND,
            Capability.SOUND_NAME,
        }
        # Custom audio is support only starting with the Windows 10 Anniversary update.
        # See https://learn.microsoft.com/en-us/windows/apps/design/shell/tiles-and-notifications/custom-audio-on-toasts#add-the-custom-audio.
        if sys.getwindowsversion().build >= 1607:  # type:ignore[attr-defined]
            capabilities.add(Capability.SOUND_FILE)

        return frozenset(capabilities)


def eventloop_wrapper(function: Callable[P, T]) -> Callable[P, None]:
    event_loop = asyncio.get_running_loop()

    def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> None:
        event_loop.call_soon_threadsafe(function, *args)

    return wrapped_func



# -*- coding: utf-8 -*-
"""
Asynchronous desktop notification API
"""
from __future__ import annotations

import asyncio
import logging
import platform
import warnings
from typing import Any, Callable, Sequence, Type, TypeVar

from packaging.version import Version

from .backends.base import DesktopNotifierBackend
from .common import (
    DEFAULT_ICON,
    DEFAULT_SOUND,
    Attachment,
    Button,
    Capability,
    Icon,
    Notification,
    ReplyField,
    Sound,
    Urgency,
)

__all__ = [
    "Notification",
    "Button",
    "ReplyField",
    "Icon",
    "Sound",
    "Attachment",
    "Urgency",
    "DesktopNotifier",
    "Capability",
    "DEFAULT_SOUND",
    "DEFAULT_ICON",
]

logger = logging.getLogger(__name__)

T = TypeVar("T")


default_event_loop_policy = asyncio.DefaultEventLoopPolicy()


def get_backend_class() -> Type[DesktopNotifierBackend]:
    """
    Return the backend class depending on the platform and version.

    :returns: A desktop notification backend suitable for the current platform.
    :raises RuntimeError: when passing ``macos_legacy = True`` on macOS 12.0 and later.
    """
    if platform.system() == "Darwin":
        from .backends.macos_support import is_bundle, is_signed_bundle, macos_version

        has_unusernotificationcenter = macos_version >= Version("10.14")

        if has_unusernotificationcenter and is_bundle():
            from .backends.macos import CocoaNotificationCenter

            if not is_signed_bundle():
                logger.warning(
                    "Could not very signature of app bundle, notifications may fail"
                )
            return CocoaNotificationCenter
        else:
            if has_unusernotificationcenter:
                logger.warning(
                    "Notification Center can only be used from an app bundle"
                )
            else:
                logger.warning("Only macOS 10.14 and later are supported")

            from .backends.dummy import DummyNotificationCenter

            return DummyNotificationCenter

    elif platform.system() == "Linux":
        from .backends.dbus import DBusDesktopNotifier

        return DBusDesktopNotifier

    elif platform.system() == "Windows" and Version(platform.version()) >= Version(
        "10.0.10240"
    ):
        from .backends.winrt import WinRTDesktopNotifier

        return WinRTDesktopNotifier

    else:
        from .backends.dummy import DummyNotificationCenter

        return DummyNotificationCenter


class DesktopNotifier:
    """Cross-platform desktop notification emitter

    Uses different backends depending on the platform version and available services.
    All implementations will dispatch notifications without an event loop but will
    require a running event loop to execute callbacks when the end user interacts with a
    notification. On Linux, an asyncio event loop is required. On macOS, a CFRunLoop *in
    the main thread* is required. Packages such as :mod:`rubicon.objc` can be used to
    integrate asyncio with a CFRunLoop.

    Callbacks to handle user interactions with a notification can be specified either at
    the class level, where they take the notification identifier as input, or directly
    on the notification itself. The latter will take precedence if set.

    Note that handlers that are directly set on the notification are tied to Python
    notification instance and therefore the app's lifecycle. Handlers that are set on
    the class level may run also for interactions with a notification while the app was
    not running, if DesktopNotifier is instantiated at app startup.

    :param app_name: Name to identify the application in the notification center. On
        Linux, this should correspond to the application name in a desktop entry. On
        macOS, this argument is ignored and the app is identified by the bundle ID of
        the sending program (e.g., Python).
    :param app_icon: Default icon to use for notifications. This should be a
        :class:`desktop_notifier.base.Icon` instance referencing either a file or a
        named system icon. :class:`str` or :class:`pathlib.Path` are also accepted but
        deprecated.
    """

    def __init__(
        self,
        app_name: str = "Python",
        app_icon: Icon | None = DEFAULT_ICON,
        notification_limit: int | None = None,
    ) -> None:
        if notification_limit is not None:
            warnings.warn(
                message="Notification limits have been deprecated and no longer have an effect",
                category=DeprecationWarning,
            )

        backend = get_backend_class()
        self._backend = backend(app_name, app_icon)
        self._did_request_authorisation = False

    @property
    def app_name(self) -> str:
        """The application name"""
        return self._backend.app_name

    @app_name.setter
    def app_name(self, value: str) -> None:
        self._backend.app_name = value

    @property
    def app_icon(self) -> Icon | None:
        """The application icon"""
        return self._backend.app_icon

    @app_icon.setter
    def app_icon(self, value: Icon | None) -> None:
        self._backend.app_icon = value

    async def request_authorisation(self) -> bool:
        """
        Requests authorisation to send user notifications. This will be automatically
        called for you when sending a notification for the first time. It also can be
        called manually to request authorisation in advance.

        On some platforms such as macOS and iOS, a prompt will be shown to the user
        when this method is called for the first time. This method does nothing on
        platforms where user authorisation is not required.

        :returns: Whether authorisation has been granted.
        """
        self._did_request_authorisation = True
        return await self._backend.request_authorisation()

    async def has_authorisation(self) -> bool:
        """Returns whether we have authorisation to send notifications."""
        return await self._backend.has_authorisation()

    async def send_notification(self, notification: Notification) -> str:
        """
        Sends a desktop notification.

        This method does not raise an exception when scheduling the notification fails
        but logs warnings instead.

        Note that even a successfully scheduled notification may not be displayed to the
        user, depending on their notification center settings (for instance if "do not
        disturb" is enabled on macOS).

        :param notification: The notification to send.
        :returns: An identifier for the scheduled notification.
        """
        # Ask for authorisation if not already done. On some platforms, this will
        # trigger a system dialog to ask the user for permission.
        if not self._did_request_authorisation:
            await self.request_authorisation()
        else:
            logger.debug("Notification center authorisation was already requested")

        # We attempt to send the notification regardless of authorization.
        # The user may have changed settings in the meantime.
        await self._backend.send(notification)

        return notification.identifier

    async def send(
        self,
        title: str,
        message: str,
        urgency: Urgency = Urgency.Normal,
        icon: Icon | None = None,
        buttons: Sequence[Button] = (),
        reply_field: ReplyField | None = None,
        on_dispatched: Callable[[], Any] | None = None,
        on_clicked: Callable[[], Any] | None = None,
        on_dismissed: Callable[[], Any] | None = None,
        attachment: Attachment | None = None,
        sound: Sound | None = None,
        thread: str | None = None,
        timeout: int = -1,  # in seconds
    ) -> str:
        """
        Sends a desktop notification

        This is a convenience function which creates a
        :class:`desktop_notifier.base.Notification` with the provided arguments and then
        calls :meth:`send_notification`.

        :returns: An identifier for the scheduled notification.
        """
        notification = Notification(
            title,
            message,
            urgency=urgency,
            icon=icon,
            buttons=tuple(buttons),
            reply_field=reply_field,
            on_dispatched=on_dispatched,
            on_clicked=on_clicked,
            on_dismissed=on_dismissed,
            attachment=attachment,
            sound=sound,
            thread=thread,
            timeout=timeout,
        )
        return await self.send_notification(notification)

    async def get_current_notifications(self) -> list[str]:
        """Returns identifiers of all currently displayed notifications for this app."""
        return await self._backend.get_current_notifications()

    async def clear(self, identifier: str) -> None:
        """
        Removes the given notification from the notification center.

        :param identifier: Notification identifier.
        """
        await self._backend.clear(identifier)

    async def clear_all(self) -> None:
        """
        Removes all currently displayed notifications for this app from the notification
        center.
        """
        await self._backend.clear_all()

    async def get_capabilities(self) -> frozenset[Capability]:
        """
        Returns which functionality is supported by the implementation.
        """
        return await self._backend.get_capabilities()

    @property
    def on_dispatched(self) -> Callable[[str], Any] | None:
        """
        A method to call when a notification is sent to the notifications server

        The method must take the notification identifier as a single argument.

        If the notification itself already specifies an on_dispatched handler, it will
        be used instead of the class-level handler.
        """
        return self._backend.on_dispatched

    @on_dispatched.setter
    def on_dispatched(self, handler: Callable[[str], Any] | None) -> None:
        self._backend.on_dispatched = handler

    @property
    def on_clicked(self) -> Callable[[str], Any] | None:
        """
        A method to call when a notification is clicked

        The method must take the notification identifier as a single argument.

        If the notification itself already specifies an on_clicked handler, it will be
        used instead of the class-level handler.
        """
        return self._backend.on_clicked

    @on_clicked.setter
    def on_clicked(self, handler: Callable[[str], Any] | None) -> None:
        self._backend.on_clicked = handler

    @property
    def on_dismissed(self) -> Callable[[str], Any] | None:
        """
        A method to call when a notification is dismissed

        The method must take the notification identifier as a single argument.

        If the notification itself already specifies an on_dismissed handler, it will be
        used instead of the class-level handler.
        """
        return self._backend.on_dismissed

    @on_dismissed.setter
    def on_dismissed(self, handler: Callable[[str], Any] | None) -> None:
        self._backend.on_dismissed = handler

    @property
    def on_button_pressed(self) -> Callable[[str, str], Any] | None:
        """
        A method to call when one of the notification's buttons is clicked

        The method must take the notification identifier and the button identifier as
        arguments.

        If the notification button itself already specifies an on_pressed handler, it
        will be used instead of the class-level handler.
        """
        return self._backend.on_button_pressed

    @on_button_pressed.setter
    def on_button_pressed(self, handler: Callable[[str, str], Any] | None) -> None:
        self._backend.on_button_pressed = handler

    @property
    def on_replied(self) -> Callable[[str, str], Any] | None:
        """
        A method to call when a user responds through the reply field of a notification

        The method must take the notification identifier and input text as arguments.

        If the notification's reply field itself already specifies an on_replied
        handler, it will be used instead of the class-level handler.
        """
        return self._backend.on_replied

    @on_replied.setter
    def on_replied(self, handler: Callable[[str, str], Any] | None) -> None:
        self._backend.on_replied = handler



# -*- coding: utf-8 -*-
"""
Synchronous desktop notification API
"""
from __future__ import annotations

import asyncio
from typing import Any, Callable, Coroutine, Sequence, TypeVar

from .common import (
    DEFAULT_ICON,
    Attachment,
    Button,
    Capability,
    Icon,
    Notification,
    ReplyField,
    Sound,
    Urgency,
)
from .main import DesktopNotifier

__all__ = ["DesktopNotifierSync"]


T = TypeVar("T")


class DesktopNotifierSync:
    """
    A synchronous counterpart to :class:`desktop_notifier.main.DesktopNotifier`

    .. warning::
        Callbacks on interaction with the notification will not work on macOS or Linux
        without a running event loop.
    """

    def __init__(
        self,
        app_name: str = "Python",
        app_icon: Icon | None = DEFAULT_ICON,
        notification_limit: int | None = None,
    ) -> None:
        self._async_api = DesktopNotifier(app_name, app_icon, notification_limit)
        self._loop = asyncio.new_event_loop()

    def _run_coro_sync(self, coro: Coroutine[None, None, T]) -> T:
        # Make sure to always use the same loop because async queues, future, etc. are
        # always bound to a loop.
        if self._loop.is_running():
            future = asyncio.run_coroutine_threadsafe(coro, self._loop)
            res = future.result()
        else:
            res = self._loop.run_until_complete(coro)

        return res

    @property
    def app_name(self) -> str:
        """The application name"""
        return self._async_api.app_name

    @app_name.setter
    def app_name(self, value: str) -> None:
        self._async_api.app_name = value

    @property
    def app_icon(self) -> Icon | None:
        """The application icon"""
        return self._async_api.app_icon

    @app_icon.setter
    def app_icon(self, value: Icon | None) -> None:
        self._async_api.app_icon = value

    def request_authorisation(self) -> bool:
        """See :meth:`desktop_notifier.main.DesktopNotifier.request_authorisation`"""
        coro = self._async_api.request_authorisation()
        return self._run_coro_sync(coro)

    def has_authorisation(self) -> bool:
        """See :meth:`desktop_notifier.main.DesktopNotifier.has_authorisation`"""
        coro = self._async_api.has_authorisation()
        return self._run_coro_sync(coro)

    def send_notification(self, notification: Notification) -> str:
        """See :meth:`desktop_notifier.main.DesktopNotifier.send_notification`"""
        coro = self._async_api.send_notification(notification)
        return self._run_coro_sync(coro)

    def send(
        self,
        title: str,
        message: str,
        urgency: Urgency = Urgency.Normal,
        icon: Icon | None = None,
        buttons: Sequence[Button] = (),
        reply_field: ReplyField | None = None,
        on_dispatched: Callable[[], Any] | None = None,
        on_clicked: Callable[[], Any] | None = None,
        on_dismissed: Callable[[], Any] | None = None,
        attachment: Attachment | None = None,
        sound: Sound | None = None,
        thread: str | None = None,
        timeout: int = -1,  # in seconds
    ) -> str:
        """See :meth:`desktop_notifier.main.DesktopNotifier.send`"""
        notification = Notification(
            title,
            message,
            urgency=urgency,
            icon=icon,
            buttons=tuple(buttons),
            reply_field=reply_field,
            on_dispatched=on_dispatched,
            on_clicked=on_clicked,
            on_dismissed=on_dismissed,
            attachment=attachment,
            sound=sound,
            thread=thread,
            timeout=timeout,
        )
        coro = self._async_api.send_notification(notification)
        return self._run_coro_sync(coro)

    def get_current_notifications(self) -> list[str]:
        """See :meth:`desktop_notifier.main.DesktopNotifier.get_current_notifications`"""
        coro = self._async_api.get_current_notifications()
        return self._run_coro_sync(coro)

    def clear(self, identifier: str) -> None:
        """See :meth:`desktop_notifier.main.DesktopNotifier.clear`"""
        coro = self._async_api.clear(identifier)
        return self._run_coro_sync(coro)

    def clear_all(self) -> None:
        """See :meth:`desktop_notifier.main.DesktopNotifier.clear_all`"""
        coro = self._async_api.clear_all()
        return self._run_coro_sync(coro)

    def get_capabilities(self) -> frozenset[Capability]:
        """See :meth:`desktop_notifier.main.DesktopNotifier.get_capabilities`"""
        coro = self._async_api.get_capabilities()
        return self._run_coro_sync(coro)
